# -*- coding: utf-8 -*-

"""The livetime module provides general functionality for detector up-time.
"""

import numpy as np

from skyllh.core.py import (
    classname,
    issequence,
)


class Livetime(
        object):
    """The ``Livetime`` class defines an interface to query the up-time of the
    detector.
    """

    @staticmethod
    def get_integrated_livetime(livetime):
        """Gets the integrated live-time from the given livetime argument, which
        is either a scalar value or an instance of Livetime.

        Parameters
        ----------
        livetime : float | Livetime instance
            The live-time in days as float, or an instance of Livetime.

        Returns
        -------
        intgrated_livetime : float
            The integrated live-time.
        """
        intgrated_livetime = livetime

        if isinstance(livetime, Livetime):
            intgrated_livetime = livetime.livetime

        return intgrated_livetime

    def __init__(
            self,
            uptime_mjd_intervals_arr,
            **kwargs):
        """Creates a new Livetime object from a (N,2)-shaped ndarray holding
        the uptime intervals.

        Parameters
        ----------
        uptime_mjd_intervals_arr : (N,2)-shaped ndarray
            The (N,2)-shaped ndarray holding the start and end times of each
            up-time interval.

            Note 1: The intervals must be sorted ascedent in time.

            Note 2: By definition the lower edge is included in the interval,
                whereas the upper edge is excluded from the interval.

            Note 3: The intervals must not overlap.

            The integrity of the internal mjd interval array will be ensured by
            the property setter method of ``uptime_mjd_intervals_arr`` by
            calling the ``assert_mjd_intervals_integrity`` method.
        """
        super().__init__(**kwargs)

        self.uptime_mjd_intervals_arr = uptime_mjd_intervals_arr

    def assert_mjd_intervals_integrity(
            self,
            arr):
        """Checks if the given MJD interval array conforms with all its
        data requirements.

        Parameters
        ----------
        arr : instance of numpy ndarray
            The (N,2)-shaped numpy ndarray holding the up-time intervals.

        Raises
        ------
        TypeError
            If the data array is not a float64 array.
        ValueError
            If the data integrity is broken.
        """
        if not isinstance(arr, np.ndarray):
            raise TypeError(
                'The internal MJD interval array must be of type ndarray! '
                'Its current type is '
                f'{classname(arr)}!')

        if arr.dtype != np.float64:
            raise TypeError(
                'The type of the internal MJD interval array is not float64!')

        # Check the shape of the array.
        if arr.ndim != 2:
            raise ValueError(
                'The dimensionality of the internel MJD interval array must '
                'be 2! Its current dimensionality is '
                f'{arr.ndim}!')
        if arr.shape[1] != 2:
            raise ValueError(
                'The length of the second axis of the internal MJD interval '
                'array must be 2! Its current length is '
                f'{arr.shape[1]}!')

        # Check if the bin edges are monotonically non decreasing.
        diff = np.diff(arr.flat)
        if not np.all(diff >= 0):
            info = ''
            for i in range(len(diff)-1):
                if diff[i] < 0:
                    info += f'i={int(i/2)}: {arr[int(i/2)]}\n'
                    info += f'i={int(i/2)+1}: {arr[int(i/2)+1]}\n'
            raise ValueError(
                'The interval edges of the internal MJD interval array are not '
                'monotonically non-decreasing!\n'
                f'{info}')

    @property
    def uptime_mjd_intervals_arr(self):
        """The Nx2 numpy ndarray holding the up-time intervals of the detector.
        The first and second elements of the second axis is the start and stop
        time of the up-time interval, respectively.
        """
        return self._uptime_mjd_intervals_arr

    @uptime_mjd_intervals_arr.setter
    def uptime_mjd_intervals_arr(self, arr):
        self.assert_mjd_intervals_integrity(arr)
        self._uptime_mjd_intervals_arr = arr

    @property
    def n_uptime_mjd_intervals(self):
        """The number of on-time intervals defined.
        """
        return self._uptime_mjd_intervals_arr.shape[0]

    @property
    def livetime(self):
        """The integrated live-time in days, based on the internal up-time time
        intervals.
        """
        return np.sum(np.diff(self._uptime_mjd_intervals_arr))

    @property
    def time_window(self):
        """(read-only) The two-element tuple holding the time window which is
        spanned by all the MJD uptime intervals.
        By definition this included possible detector down-time periods.
        """
        return (self._uptime_mjd_intervals_arr[0, 0],
                self._uptime_mjd_intervals_arr[-1, 1])

    @property
    def time_start(self):
        """(read-only) The start time of the detector live-time.
        """
        return self._uptime_mjd_intervals_arr[0, 0]

    @property
    def time_stop(self):
        """(read-only) The stop time of the detector live-time.
        """
        return self._uptime_mjd_intervals_arr[-1, 1]

    def __str__(self):
        """Pretty string representation of the Livetime class instance.
        """
        s = (f'{classname(self)}(time_window=('
             f'{self.time_window[0]:.6f}, {self.time_window[1]:.6f}))')
        return s

    def _get_onoff_intervals(self):
        """A view on the uptime intervals where each time is a lower bin edge.
        Hence, odd array elements (bins) are on-time intervals, and even array
        elements are off-time intervals.

        Returns
        -------
        onoff_intervals : instance of numpy ndarray
            The (n_uptime_intervals*2,)-shaped numpy ndarray holding the time
            edges of the uptime intervals.
        """
        onoff_intervals = np.reshape(
            self._uptime_mjd_intervals_arr,
            (self._uptime_mjd_intervals_arr.size,))

        return onoff_intervals

    def _get_onoff_interval_indices(self, mjds):
        """Retrieves the indices of the on-time and off-time intervals, which
        correspond to the given MJD values.

        Odd indices correspond to detector on-time intervals and even indices
        to detector off-time intervals.

        The indices are in the range (0, 2*N), where N is the number of on-time
        intervals. The index 0 corresponds to a time prior to the first on-time
        interval, whereas 2*N corresponds to a time past the last on-time
        interval.

        Parameters
        ----------
        mjds : numpy array of floats
            The array of MJD values.

        Returns
        -------
        idxs : numpy array of ints
            The array of the on-off-time interval indices that correspond to the
            given MJD values.
        """
        # Get the interval indices.
        # Note: For MJD values outside the total interval range, the np.digitize
        # function will return either 0, or len(bins). Since, there is always
        # an even amount of intervals edges, and 0 is also an 'even' number,
        # those MJDs will correspond to off-time automatically.
        idxs = np.digitize(mjds, self._get_onoff_intervals())

        return idxs

    def get_uptime_intervals_between(
            self,
            t_start,
            t_end):
        """Creates a (N,2)-shaped ndarray holding the on-time detector intervals
        between the given time range from t_start to t_end.

        Parameters
        ----------
        t_start : float
            The MJD start time of the time range to consider. This might be the
            lower bound of the first on-time interval.
        t_end : float
            The MJD end time of the time range to consider. This might be the
            upper bound of the last on-time interval.

        Returns
        -------
        ontime_intervals : (N,2)-shaped ndarray
            The (N,2)-shaped ndarray holding the on-time detector intervals.
        """
        onoff_intervals = self._get_onoff_intervals()

        (t_start_idx, t_end_idx) = self._get_onoff_interval_indices(
            (t_start, t_end))
        if t_start_idx % 2 == 0:
            # t_start is during off-time. Use the next on-time lower edge as
            # first on-time edge.
            t_start = onoff_intervals[t_start_idx]
        else:
            t_start_idx -= 1
        if t_end_idx % 2 == 0:
            # t_end is during off-time. Use the previous on-time upper edge as
            # the last on-time edge.
            t_end = onoff_intervals[t_end_idx-1]
        else:
            t_end_idx += 1

        # The t_start_idx and t_end_idx variables hold even indices.
        N_ontime_intervals = int((t_end_idx - t_start_idx)/2)

        ontime_intervals_flat = np.empty(
            (N_ontime_intervals*2,), dtype=np.float64)
        # Set the first and last on-time interval edges.
        ontime_intervals_flat[0] = t_start
        ontime_intervals_flat[-1] = t_end
        if N_ontime_intervals > 1:
            # Fill also the interval edges of the intermediate on-time bins.
            ontime_intervals_flat[1:-1] = onoff_intervals[t_start_idx+1:t_end_idx-1]

        ontime_intervals = np.reshape(
            ontime_intervals_flat,
            (N_ontime_intervals, 2))

        return ontime_intervals

    def get_livetime_upto(self, mjd):
        """Calculates the cumulative detector livetime up to the given time.

        Parameters
        ----------
        mjd : float | array of floats
            The time in MJD up to which the detector livetime should be
            calculated.

        Returns
        -------
        livetimes : float | ndarray of floats
            The ndarray holding the cumulative detector livetime corresponding
            to the the given MJD times.
        """
        mjds = np.atleast_1d(mjd)

        onoff_idxs = self._get_onoff_interval_indices(mjds)

        # Create a mask for all the odd indices, i.e. the MJDs falling into an
        # on-time interval.
        odd_idxs_mask = onoff_idxs & 0x1

        # Map the indices to the cum_ontime_bins array. Off-time indices will
        # be mapped to its prior on-time interval.
        #                               Odd indices.     Even indices.
        idxs = np.where(odd_idxs_mask, (onoff_idxs-1)/2, onoff_idxs/2 - 1)
        # At this point, there could be indices of value -1 from MJD values
        # prior to the first on-time interval. So we just move all the indices
        # by one.
        idxs += 1

        # Create a cumulative on-time array with a leading 0 element for MJDs
        # prior to the first on-time interval.
        ontime_bins = np.diff(self._uptime_mjd_intervals_arr).reshape(
            (self.n_uptime_mjd_intervals,))
        cum_ontime_bins = np.array([0], dtype=np.float64)
        cum_ontime_bins = np.append(cum_ontime_bins, np.cumsum(ontime_bins))

        # For odd (on-time) mjds, use the cumulative value of the previous bin
        # and add the part of the interval bin up to the mjd value.
        livetimes = np.where(
            odd_idxs_mask,
            cum_ontime_bins[idxs-1]
            + mjds
            - self._get_onoff_intervals()[onoff_idxs-1],
            cum_ontime_bins[idxs])

        if not issequence(mjd):
            return np.asscalar(livetimes)

        return livetimes

    def is_on(self, mjd):
        """Checks if the detector is on at the given MJD time. MJD times
        outside any live-time interval will be masked as False.

        Parameters
        ----------
        mjd : float | sequence of float
            The time in MJD.

        Returns
        -------
        is_on : array of bool
            True if the detector was on at the given time.
        """
        mjd = np.atleast_1d(mjd)

        # Get the on-off-time interval indices corresponding to the given MJD
        # values.
        onoff_idxs = self._get_onoff_interval_indices(mjd)

        # Mask odd indices as on-time (True) MJD values and even indices as
        # off-time (False).
        is_on = np.array(onoff_idxs & 0x1, dtype=np.bool_)

        return is_on

    def draw_ontimes(
            self,
            rss,
            size,
            t_min=None,
            t_max=None):
        """Draws random MJD times based on the detector on-time intervals.

        Parameters
        ----------
        rss : RandomStateService
            The skyllh RandomStateService instance to use for drawing random
            numbers from.
        size : int
            The number of random MJD times to generate.
        t_min : float
            The optional minimal time to consider. If set to ``None``, the
            start time of this Livetime instance will be used.
        t_max : float
            The optional maximal time to consider. If set to ``None``, the
            end time of this Livetime instance will be used.

        Returns
        -------
        ontimes : ndarray
            The 1d array holding the generated MJD times.
        """
        uptime_intervals_arr = self._uptime_mjd_intervals_arr

        if t_min is not None or t_max is not None:
            if t_min is None:
                t_min = self.time_start
            if t_max is None:
                t_max = self.time_stop

            uptime_intervals_arr = self.get_uptime_intervals_between(
                t_min, t_max)

        onoff_intervals = np.reshape(
            uptime_intervals_arr,
            (uptime_intervals_arr.size,))

        # Create bin array with only on-time bins. We have to mask out the
        # off-time bins.
        ontime_bins = np.diff(onoff_intervals)
        mask = np.invert(
            np.array(
                np.linspace(0, ontime_bins.size-1, ontime_bins.size) % 2,
                dtype=np.bool_))
        ontime_bins = ontime_bins[mask]

        # Create the cumulative array of the on-time bins.
        cum_ontime_bins = np.array([0], dtype=np.float64)
        cum_ontime_bins = np.append(cum_ontime_bins, np.cumsum(ontime_bins))

        #         |<--y->|
        # |----|  |-----------|  |-------|
        # l1   u1 l2     |xL  u2 ul3     u3
        #
        # x \el [0,1]
        # L = \sum (u_i - l_i)

        x = rss.random.uniform(0, 1, size)
        # Get the sum L of all the on-time intervals.
        L = cum_ontime_bins[-1]
        w = x*L
        idxs = np.digitize(w, cum_ontime_bins)
        lower = uptime_intervals_arr[:, 0]
        y = w - cum_ontime_bins[idxs-1]
        ontimes = lower[idxs-1] + y

        return ontimes



# -*- coding: utf-8 -*-

import numpy as np

from scipy import (
    interpolate,
)

from skyllh.analyses.i3.publicdata_ps.aeff import (
    PDAeff,
)
from skyllh.analyses.i3.publicdata_ps.smearing_matrix import (
    PDSmearingMatrix,
)
from skyllh.analyses.i3.publicdata_ps.utils import (
    psi_to_dec_and_ra,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.flux_model import (
    TimeFluxProfile,
)
from skyllh.core.livetime import (
    Livetime,
)
from skyllh.core.py import (
    classname,
    float_cast,
    int_cast,
    module_classname,
)
from skyllh.core.signal_generator import (
    SignalGenerator,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
)
from skyllh.core.utils.flux_model import (
    create_scipy_stats_rv_continuous_from_TimeFluxProfile,
)


class PDDatasetSignalGenerator(
        SignalGenerator,
):
    """This class implements a signal generator for a single public data
    dataset.
    """
    def __init__(
            self,
            shg_mgr,
            ds,
            ds_idx,
            energy_cut_spline=None,
            cut_sindec=None,
            **kwargs,
    ):
        """Creates a new instance of the signal generator for generating
        signal events from a specific public data dataset.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager defining the source
            hypothesis groups.
        ds : instance of Dataset
            The instance of Dataset for which signal events should get
            generated.
        ds_idx : int
            The index of the dataset.
        energy_cut_spline : scipy.interpolate.UnivariateSpline
            A spline of E(sin_dec) that defines the declination
            dependent energy cut in the IceCube southern sky.
        cut_sindec : float
            The sine of the declination to start applying the energy cut.
            The cut will be applied from this declination down.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            **kwargs)

        self._logger = get_logger(module_classname(self))

        self.ds = ds
        self.ds_idx = ds_idx
        self.energy_cut_spline = energy_cut_spline
        self.cut_sindec = cut_sindec

        self.sm = PDSmearingMatrix(
            pathfilenames=ds.get_abs_pathfilename_list(
                    ds.get_aux_data_definition('smearing_datafile')))

        self._create_source_dependent_data_structures()

    def _create_source_dependent_data_structures(self):
        """Creates the source dependent data structures needed by this signal
        generator. These are:

            - source location in ra and dec
            - effective area
            - log10 true energy inv CDF spline

        """
        n_sources = self.shg_mgr.n_sources

        self._src_ra_arr = np.empty(
            (n_sources,),
            dtype=np.float64)
        self._src_dec_arr = np.empty(
            (n_sources,),
            dtype=np.float64)
        self._effA_arr = np.empty(
            (n_sources,),
            dtype=np.object_)
        self._log10_true_e_inv_cdf_spl_arr = np.empty(
            (n_sources,),
            dtype=np.object_)

        for (src_idx, src) in enumerate(self._shg_mgr.source_list):
            self._src_ra_arr[src_idx] = src.ra
            self._src_dec_arr[src_idx] = src.dec

            dec_idx = self.sm.get_true_dec_idx(src.dec)
            (min_log_true_e,
             max_log_true_e) =\
                self.sm.get_true_log_e_range_with_valid_log_e_pdfs(
                    dec_idx)

            self._effA_arr[src_idx] = PDAeff(
                pathfilenames=self.ds.get_abs_pathfilename_list(
                    self.ds.get_aux_data_definition('eff_area_datafile')),
                src_dec=src.dec,
                min_log10enu=min_log_true_e,
                max_log10enu=max_log_true_e)

            # Build the spline for the inverse CDF of the source flux's true
            # energy probability distribution.
            fluxmodel = self.shg_mgr.get_fluxmodel_by_src_idx(src_idx=src_idx)
            self._log10_true_e_inv_cdf_spl_arr[src_idx] =\
                self._create_inv_cdf_spline(
                    src_idx=src_idx,
                    fluxmodel=fluxmodel,
                    log_e_min=min_log_true_e,
                    log_e_max=max_log_true_e)

    @staticmethod
    def _eval_spline(x, spl):
        """Evaluates the given spline at the given coordinates.
        """
        x = np.asarray(x)
        if (x.any() < 0) or (x.any() > 1):
            raise ValueError(
                f'{x} is outside of the valid spline range. '
                'The valid range is [0,1].')

        values = interpolate.splev(x, spl, ext=3)

        return values

    def _create_inv_cdf_spline(
            self,
            src_idx,
            fluxmodel,
            log_e_min,
            log_e_max):
        """Creates a spline for the inverse cumulative distribution function of
        the detectable true energy probability distribution.
        """
        effA = self._effA_arr[src_idx]

        m = (effA.log10_enu_bincenters >= log_e_min) & (
             effA.log10_enu_bincenters < log_e_max)
        bin_centers = effA.log10_enu_bincenters[m]
        low_bin_edges = effA.log10_enu_binedges_lower[m]
        high_bin_edges = effA.log10_enu_binedges_upper[m]

        # Flux probability P(E_nu | gamma) per bin.
        flux_prob = fluxmodel.energy_profile.get_integral(
            E1=10**low_bin_edges,
            E2=10**high_bin_edges
        ) / fluxmodel.energy_profile.get_integral(
            E1=10**low_bin_edges[0],
            E2=10**high_bin_edges[-1]
        )

        # Do the product and normalize again to a probability per bin.
        product = flux_prob * effA.det_prob
        prob_per_bin = product / np.sum(product)

        # The probability per bin cannot be zero, otherwise the cumulative
        # sum would not be increasing monotonically. So we set zero bins to
        # 1000 times smaller than the smallest non-zero bin.
        m = prob_per_bin == 0
        prob_per_bin[m] = np.min(prob_per_bin[np.invert(m)]) / 1000
        to_keep = prob_per_bin > 1e-15  # For numerical stability.
        prob_per_bin = prob_per_bin[to_keep]
        prob_per_bin /= np.sum(prob_per_bin)

        # Compute the cumulative distribution CDF.
        cum_per_bin = [
            np.sum(prob_per_bin[:i])
            for i in range(prob_per_bin.size+1)
        ]
        if np.any(np.diff(cum_per_bin) == 0):
            raise ValueError(
                'The cumulative sum of the true energy probability is not '
                'monotonically increasing! Values of the cumsum are '
                f'{cum_per_bin}.')

        bin_centers = bin_centers[to_keep]
        bin_centers = np.concatenate(([low_bin_edges[0]], bin_centers))

        # Build a spline for the inverse CDF.
        return interpolate.splrep(cum_per_bin, bin_centers, k=1, s=0)

    def _draw_signal_events_for_source(
            self,
            rss,
            src_dec,
            src_ra,
            dec_idx,
            log10_true_e_inv_cdf_spl,
            n_events):
        """Generates `n_events` signal events for the given source location
        given the given inverse cumulative density function for the
        log10(E_true/GeV) distribution.

        Note:
            Some values can be NaN in cases where a PDF was not available!

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService to use for drawing random
            numbers.
        src_dec : float
            The declination of the source in radians.
        src_ra : float
            The right-ascention of the source in radians.
        dec_idx : int
            The SM's declination bin index of the source's declination.
        log10_true_e_inv_cdf_spl : instance of scipy.interpolate.splrep
            The linear spline interpolation representation of the inverse
            cummulative density function of the log10(E_true/GeV) distribution.
        n_events : int
            The number of events to generate.

        Returns
        -------
        events : instance of DataFieldRecordArray of size `n_events`
            The instance of DataFieldRecordArray of length `n_events` holding
            the event data. It contains the following data fields:

                ``'isvalid'``
                ``'log_true_energy'``
                ``'log_energy'``
                ``'dec'``
                ``'sin_dec'``
                ``'ang_err'``
                ``'time'``
                ``'azi'``
                ``'zen'``
                ``'run'``

            Single values can be NaN in cases where a pdf was not available.
        """
        # Create the output event DataFieldRecordArray.
        out_dtype = [
            ('isvalid', np.bool_),
            ('log_true_energy', np.double),
            ('log_energy', np.double),
            ('dec', np.double),
            ('ra', np.double),
            ('sin_dec', np.double),
            ('ang_err', np.double),
            ('time', int),
            ('azi', np.double),
            ('zen', np.double),
            ('run', int)
        ]

        data = dict(
            [(out_dt[0], np.empty(
                (n_events,),
                dtype=out_dt[1])
              ) for out_dt in out_dtype]
        )

        events = DataFieldRecordArray(data, copy=False)

        sm = self.sm

        log_true_e = self._eval_spline(
            rss.random.uniform(size=n_events), log10_true_e_inv_cdf_spl)

        events['log_true_energy'] = log_true_e

        log_true_e_idxs = (
            np.digitize(log_true_e, bins=sm.true_e_bin_edges) - 1
        )

        # Sample reconstructed energies given true neutrino energies.
        (log_e_idxs, log_e) = sm.sample_log_e(rss, dec_idx, log_true_e_idxs)
        events['log_energy'] = log_e

        # Sample reconstructed psi values given true neutrino energy and
        # reconstructed energy.
        (psi_idxs, psi) = sm.sample_psi(
            rss, dec_idx, log_true_e_idxs, log_e_idxs)

        # Sample reconstructed ang_err values given true neutrino energy,
        # reconstructed energy, and psi.
        (ang_err_idxs, ang_err) = sm.sample_ang_err(
            rss, dec_idx, log_true_e_idxs, log_e_idxs, psi_idxs)

        isvalid = np.invert(
            np.isnan(log_e) | np.isnan(psi) | np.isnan(ang_err))
        events['isvalid'] = isvalid

        # Convert the psf into a set of (r.a. and dec.). Only use non-nan
        # values.
        (dec, ra) = psi_to_dec_and_ra(rss, src_dec, src_ra, psi[isvalid])
        events['ra'][isvalid] = ra
        events['dec'][isvalid] = dec
        events['sin_dec'][isvalid] = np.sin(dec)

        # Add an angular error. Only use non-nan values.
        events['ang_err'][isvalid] = ang_err[isvalid]

        # Add fields required by the framework.
        events['time'] = np.full((n_events,), np.nan, dtype=np.float64)
        events['azi'] = np.full((n_events,), np.nan, dtype=np.float64)
        events['zen'] = np.full((n_events,), np.nan, dtype=np.float64)
        events['run'] = np.full((n_events,), -1, dtype=np.int64)

        return events

    def change_shg_mgr(
            self,
            shg_mgr):
        """Changes the source hypothesis group manager. This will recreate the
        internal source dependent data structures.
        """
        super().change_shg_mgr(
            shg_mgr=shg_mgr)

        self._create_source_dependent_data_structures()

    @staticmethod
    def create_energy_filter_mask(
            events,
            spline,
            cut_sindec,
            logger):
        """Creates a mask for cutting all events below ``cut_sindec``
        that have an energy smaller than the energy spline at their
        declination.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated signal
            events.
        spline : instance of scipy.interpolate.UnivariateSpline
            A spline of E(sin_dec) that defines the declination
            dependent energy cut in the IceCube southern sky.
        cut_sindec : float
            The sine of the declination to start applying the energy cut.
            The cut will be applied from this declination down.
        logger : instance of logging.Logger
            The Logger instance.

        Returns
        -------
        filter_mask : instance of numpy ndarray
            The (len(events),)-shaped numpy ndarray with the mask of the events
            to cut.
        """
        if cut_sindec is None:
            logger.warn(
                'No `cut_sindec` has been specified. The energy cut will be '
                'applied in [-90, 0] deg.')
            cut_sindec = 0.

        filter_mask = np.logical_and(
            events['sin_dec'] < cut_sindec,
            events['log_energy'] < spline(events['sin_dec']))

        return filter_mask

    def generate_signal_events_for_source(
            self,
            rss,
            src_idx,
            n_events):
        """Generates ``n_events`` signal events for the given source location
        and flux model.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService providing the random number
            generator state.
        src_idx : int
            The index of the source.
        n_events : int
            Number of signal events to be generated.

        Returns
        -------
        events : instance of DataFieldRecordArray
            The numpy record array holding the event data.
            It contains the following data fields:

                - 'isvalid'
                - 'log_true_energy'
                - 'log_energy'
                - 'dec'
                - 'ra'
                - 'ang_err'

        """
        sm = self.sm

        src_dec = self._src_dec_arr[src_idx]
        src_ra = self._src_ra_arr[src_idx]

        log10_true_e_inv_cdf_spl = self._log10_true_e_inv_cdf_spl_arr[src_idx]

        # Find the declination bin index.
        dec_idx = sm.get_true_dec_idx(src_dec)

        events = None

        n_events_generated = 0
        while n_events_generated < n_events:
            n_evt = n_events - n_events_generated

            events_ = self._draw_signal_events_for_source(
                rss=rss,
                src_dec=src_dec,
                src_ra=src_ra,
                dec_idx=dec_idx,
                log10_true_e_inv_cdf_spl=log10_true_e_inv_cdf_spl,
                n_events=n_evt)

            # Cut events that failed to be generated due to missing PDFs.
            # Also cut low energy events if generating in the southern sky.
            events_ = events_[events_['isvalid']]

            if self.energy_cut_spline is not None:
                cut_mask = self.create_energy_filter_mask(
                    events=events_,
                    spline=self.energy_cut_spline,
                    cut_sindec=self.cut_sindec,
                    logger=self._logger)
                events_ = events_[~cut_mask]

            if len(events_) > 0:
                n_events_generated += len(events_)
                if events is None:
                    events = events_
                else:
                    events.append(events_)

        return events

    def generate_signal_events(
            self,
            rss,
            mean,
            poisson=True,
            src_detsigyield_weights_service=None,
            **kwargs):
        """Generates ``mean`` number of signal events.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService providing the random number
            generator state.
        mean : int | float
            The mean number of signal events. If the ``poisson`` argument is set
            to True, the actual number of generated signal events will be drawn
            from a Poisson distribution with this given mean value of signal
            events.
        poisson : bool
            If set to True, the actual number of generated signal events will
            be drawn from a Poisson distribution with the given mean value of
            signal events.
            If set to False, the argument ``mean`` specifies the actual number
            of generated signal events.
        src_detsigyield_weights_service : instance of SrcDetSigYieldWeightsService
            The instance of SrcDetSigYieldWeightsService providing the weighting
            of the sources within the detector.

        Returns
        -------
        n_signal : int
            The number of generated signal events.
        signal_events_dict : dict of DataFieldRecordArray
            The dictionary holding the DataFieldRecordArray instances with the
            generated signal events. Each key of this dictionary represents the
            dataset index for which the signal events have been generated.
        """
        if poisson:
            n_events = rss.random.poisson(
                float_cast(
                    mean,
                    'The `mean` argument must be castable to type of float!'))

        n_events = int_cast(
            mean,
            'The `mean` argument must be castable to type of int!')

        if src_detsigyield_weights_service is None:
            raise ValueError(
                'The src_detsigyield_weights_service argument must be provided '
                f'for the signal generator {classname(self)}!')

        (a_jk, a_jk_grads) = src_detsigyield_weights_service.get_weights()

        a_k = np.copy(a_jk[self.ds_idx])
        a_k /= np.sum(a_k)

        n_signal = 0
        signal_events_dict = {}

        # Loop over the sources and generate signal events according to the
        # weights of the sources.
        for src_idx in range(self.shg_mgr.n_sources):
            n_events_src = int(np.round(n_events * a_k[src_idx], 0))

            src_events = self.generate_signal_events_for_source(
                rss=rss,
                src_idx=src_idx,
                n_events=n_events_src,
            )
            if src_events is None:
                continue

            n_signal += len(src_events)

            if self.ds_idx not in signal_events_dict:
                signal_events_dict[self.ds_idx] = src_events
            else:
                signal_events_dict[self.ds_idx].append(src_events)

        return (n_signal, signal_events_dict)


class TimeDependentPDDatasetSignalGenerator(
        PDDatasetSignalGenerator,
):
    """This time dependent signal generator for a public PS dataset generates
    events using the
    :class:`~skyllh.analyses.i3.publicdata_ps.signal_generator.PDDatasetSignalGenerator`
    class. It then draws times for each event and adds them to the event array.
    """

    def __init__(
            self,
            shg_mgr,
            ds,
            ds_idx,
            livetime,
            time_flux_profile,
            energy_cut_spline=None,
            cut_sindec=None,
            **kwargs,
    ):
        """
        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            source hypothesis groups, i.e. the list of sources.
        ds : instance of Dataset
            The instance of Dataset for which signal events should get
            generated.
        ds_idx : int
            The index of the dataset.
        livetime : instance of Livetime
            The instance of Livetime providing the live-time information of the
            dataset.
        time_flux_profile : instance of TimeFluxProfile
            The instance of TimeFluxProfile providing the time profile of the
            source(s).

            Note:

                At this time the some time profile will be used for all
                sources!

        energy_cut_spline : scipy.interpolate.UnivariateSpline
            A spline of E(sin_dec) that defines the declination
            dependent energy cut in the IceCube southern sky.
        cut_sindec : float
            The sine of the declination to start applying the energy cut.
            The cut will be applied from this declination down.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            ds=ds,
            ds_idx=ds_idx,
            energy_cut_spline=energy_cut_spline,
            cut_sindec=cut_sindec,
            **kwargs)

        if not isinstance(time_flux_profile, TimeFluxProfile):
            raise TypeError(
                'The time_flux_profile argument must be an instance of '
                'TimeFluxProfile! '
                f'Its current type is {classname(time_flux_profile)}!')

        self.livetime = livetime
        self._time_flux_profile = time_flux_profile

    @property
    def livetime(self):
        """The instance of Livetime providing the live-time information of the
        dataset.
        """
        return self._livetime

    @livetime.setter
    def livetime(self, lt):
        if not isinstance(lt, Livetime):
            raise TypeError(
                'The livetime property must be an instance of Livetime! '
                f'Its current type is {classname(lt)}!')
        self._livetime = lt

    def generate_signal_events(
            self,
            rss,
            mean,
            poisson=True,
            src_detsigyield_weights_service=None,
            **kwargs,
    ):
        """Generates ``mean`` number of signal events with times.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService providing the random number
            generator state.
        mean : int | float
            The mean number of signal events. If the ``poisson`` argument is set
            to True, the actual number of generated signal events will be drawn
            from a Poisson distribution with this given mean value of signal
            events.
        poisson : bool
            If set to True, the actual number of generated signal events will
            be drawn from a Poisson distribution with the given mean value of
            signal events.
            If set to False, the argument ``mean`` specifies the actual number
            of generated signal events.
        src_detsigyield_weights_service : instance of SrcDetSigYieldWeightsService
            The instance of SrcDetSigYieldWeightsService providing the weighting
            of the sources within the detector.

        Returns
        -------
        n_signal : int
            The number of generated signal events.
        signal_events_dict : dict of DataFieldRecordArray
            The dictionary holding the DataFieldRecordArray instances with the
            generated signal events. Each key of this dictionary represents the
            dataset index for which the signal events have been generated.
        """
        (n_signal, signal_events_dict) = super().generate_signal_events(
            rss=rss,
            mean=mean,
            poisson=poisson,
            src_detsigyield_weights_service=src_detsigyield_weights_service,
            **kwargs)

        # Create a scipy.stats.rv_continuous instance for the time flux profile.
        time_rv = create_scipy_stats_rv_continuous_from_TimeFluxProfile(
            profile=self._time_flux_profile)

        # Optimized time injection version, based on csky implementation.
        # https://github.com/icecube/csky/blob/7e969639c5ef6dbb42872dac9b761e1e8b0ccbe2/csky/inj.py#L1122
        events = signal_events_dict[self.ds_idx]
        times = np.array([], dtype=np.float64)
        n_events = len(events)
        while len(times) < n_events:
            new_times = time_rv.rvs(
                size=(n_events - len(times)),
                random_state=rss.random)
            mask = self._livetime.is_on(
                mjd=new_times)
            new_times = new_times[mask]

            times = np.concatenate((times, new_times))
        events['time'] = times

        return (n_signal, signal_events_dict)



# -*- coding: utf-8 -*-

import scipy.interpolate

import numpy as np

from skyllh.core.binning import (
    UsesBinning,
)
from skyllh.core.pdf import (
    IsBackgroundPDF,
    SpatialPDF,
)
from skyllh.core.py import (
    classname,
    int_cast,
    issequence,
    issequenceof,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
)
from skyllh.core.timing import (
    TaskTimer,
)
from skyllh.i3.pdf import (
    I3EnergyPDF,
)


class BackgroundI3SpatialPDF(
        SpatialPDF,
        UsesBinning,
        IsBackgroundPDF,
):
    """This is the base class for all IceCube specific spatial background PDF
    models. IceCube spatial background PDFs depend solely on the zenith angle,
    and hence, on the declination of the event.

    The IceCube spatial background PDF is modeled as a 1d spline function in
    sin(declination).
    """
    def __init__(
            self,
            data_sin_dec,
            data_weights,
            sin_dec_binning,
            spline_order_sin_dec,
            **kwargs,
    ):
        """Creates a new IceCube spatial background PDF object.

        Parameters
        ----------
        data_sin_dec : 1d ndarray
            The array holding the sin(dec) values of the events.
        data_weights : 1d ndarray
            The array holding the weight of each event used for histogramming.
        sin_dec_binning : BinningDefinition
            The binning definition for the sin(declination) axis.
        spline_order_sin_dec : int
            The order of the spline function for the logarithmic values of the
            spatial background PDF along the sin(dec) axis.
        """
        super().__init__(
            pmm=None,
            ra_range=(0, 2*np.pi),
            dec_range=(
                np.arcsin(sin_dec_binning.lower_edge),
                np.arcsin(sin_dec_binning.upper_edge)),
            **kwargs)

        self.add_binning(sin_dec_binning, 'sin_dec')
        self.spline_order_sin_dec = spline_order_sin_dec

        (h, bins) = np.histogram(
            data_sin_dec,
            bins=sin_dec_binning.binedges,
            weights=data_weights,
            range=sin_dec_binning.range)

        # Save original histogram.
        self._orig_hist = h

        # Normalize histogram to get PDF.
        h = h / h.sum() / (bins[1:] - bins[:-1])

        # Check if there are any NaN values.
        if np.any(np.isnan(h)):
            nan_bcs = sin_dec_binning.bincenters[np.isnan(h)]
            raise ValueError(
                'The declination histogram contains NaN values! Check your '
                'sin(dec) binning! The bins with NaN values are: '
                f'{nan_bcs}')

        if np.any(h <= 0.):
            empty_bcs = sin_dec_binning.bincenters[h <= 0.]
            raise ValueError(
                'Some declination histogram bins for the spatial background '
                'PDF are empty, this must not happen! The empty bins are: '
                f'{empty_bcs}')

        # Create the logarithmic spline.
        self._log_spline = scipy.interpolate.InterpolatedUnivariateSpline(
            sin_dec_binning.bincenters, np.log(h), k=self.spline_order_sin_dec)

        # Save original spline.
        self._orig_log_spline = self._log_spline

    @property
    def spline_order_sin_dec(self):
        """The order (int) of the logarithmic spline function, that splines the
        background PDF, along the sin(dec) axis.
        """
        return self._spline_order_sin_dec

    @spline_order_sin_dec.setter
    def spline_order_sin_dec(self, order):
        self._spline_order_sin_dec = int_cast(
            order,
            'The spline_order_sin_dec property must be castable to type int!')

    def add_events(self, events):
        """Add events to spatial background PDF object and recalculate
        logarithmic spline function.

        Parameters
        ----------
        events : numpy record ndarray
            The array holding the event data. The following data fields must
            exist:

                sin_dec : float
                    The sin(declination) value of the event.

        """
        data_sin_dec = events['sin_dec']

        sin_dec_binning = self.get_binning('sin_dec')

        (h_upd, bins) = np.histogram(
            data_sin_dec,
            bins=sin_dec_binning.binedges,
            range=sin_dec_binning.range)

        # Construct histogram with added events.
        h = self._orig_hist + h_upd

        # Normalize histogram to get PDF.
        h = h / h.sum() / (bins[1:] - bins[:-1])

        # Create the updated logarithmic spline.
        self._log_spline = scipy.interpolate.InterpolatedUnivariateSpline(
            sin_dec_binning.bincenters, np.log(h), k=self.spline_order_sin_dec)

    def reset(self):
        """Reset the logarithmic spline to the original function, which was
        calculated when the object was initialized.
        """
        self._log_spline = self._orig_log_spline

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs):
        """Pre-cumputes the probability density values when new trial data is
        available.
        """
        with TaskTimer(tl, 'Evaluating bkg log-spline.'):
            log_spline_val = self._log_spline(tdm.get_data('sin_dec'))

        self._pd = 0.5 / np.pi * np.exp(log_spline_val)

    def get_pd(
            self,
            tdm,
            params_recarray=None,
            tl=None):
        """Calculates the spatial background probability on the sphere of each
        event.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial event data for which
            to calculate the PDF values. The following data fields must exist:

                sin_dec : float
                    The sin(declination) value of the event.

        params_recarray : None
            Unused interface parameter.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of numpy ndarray
            The (N_events,)-shaped numpy ndarray holding the background
            probability density value for each event.
        grads : dict
            The dictionary holding the gradients of the probability density
            w.r.t. each global fit parameter.
            The background PDF does not depend on any global fit parameter,
            hence, this is an empty dictionary.
        """
        return (self._pd, dict())


class DataBackgroundI3SpatialPDF(
        BackgroundI3SpatialPDF,
):
    """This is the IceCube spatial background PDF, which gets constructed from
    experimental data.
    """
    def __init__(
            self,
            data_exp,
            sin_dec_binning,
            spline_order_sin_dec=2,
            **kwargs,
    ):
        """Constructs a new IceCube spatial background PDF from experimental
        data.

        Parameters
        ----------
        data_exp : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the experimental data.
            The following data fields must exist:

                sin_dec : float
                    The sin(declination) of the data event.

        sin_dec_binning : BinningDefinition
            The binning definition for the sin(declination).
        spline_order_sin_dec : int
            The order of the spline function for the logarithmic values of the
            spatial background PDF along the sin(dec) axis.
            The default is 2.
        """
        if not isinstance(data_exp, DataFieldRecordArray):
            raise TypeError(
                'The data_exp argument must be an instance of '
                'DataFieldRecordArray! '
                f'It is of type "{classname(data_exp)}"!')

        data_sin_dec = data_exp['sin_dec']
        data_weights = np.ones((len(data_exp),))

        # Create the PDF using the base class.
        super().__init__(
            data_sin_dec=data_sin_dec,
            data_weights=data_weights,
            sin_dec_binning=sin_dec_binning,
            spline_order_sin_dec=spline_order_sin_dec,
            **kwargs)


class MCBackgroundI3SpatialPDF(
        BackgroundI3SpatialPDF,
):
    """This is the IceCube spatial background PDF, which gets constructed from
    monte-carlo data.
    """
    def __init__(
            self,
            data_mc,
            physics_weight_field_names,
            sin_dec_binning,
            spline_order_sin_dec=2,
            **kwargs,
    ):
        """Constructs a new IceCube spatial background PDF from monte-carlo
        data.

        Parameters
        ----------
        data_mc : instance of DataFieldRecordArray
            The array holding the monte-carlo data. The following data fields
            must exist:

                sin_dec : float
                    The sine of the reconstructed declination of the data event.

        physics_weight_field_names : str | list of str
            The name or the list of names of the monte-carlo data fields, which
            should be used as event weights. If a list is given, the weight
            values of all the fields will be summed to construct the final event
            weight.
        sin_dec_binning : BinningDefinition
            The binning definition for the sin(declination).
        spline_order_sin_dec : int
            The order of the spline function for the logarithmic values of the
            spatial background PDF along the sin(dec) axis.
            The default is 2.
        """
        if not isinstance(data_mc, DataFieldRecordArray):
            raise TypeError(
                'The data_mc argument must be and instance of '
                'DataFieldRecordArray! '
                f'It is of type {classname(data_mc)}')

        if not issequence(physics_weight_field_names):
            physics_weight_field_names = [physics_weight_field_names]
        if not issequenceof(physics_weight_field_names, str):
            raise TypeError(
                'The physics_weight_field_names argument must be of type str '
                'or a sequence of type str! It is of type '
                f'"{classname(physics_weight_field_names)}"!')

        data_sin_dec = data_mc['sin_dec']

        # Calculate the event weights as the sum of all the given data fields
        # for each event.
        data_weights = np.zeros(len(data_mc), dtype=np.float64)
        for name in physics_weight_field_names:
            if name not in data_mc:
                raise KeyError(
                    f'The field "{name}" does not exist in the MC data!')
            data_weights += data_mc[name]

        # Create the PDF using the base class.
        super().__init__(
            data_sin_dec=data_sin_dec,
            data_weights=data_weights,
            sin_dec_binning=sin_dec_binning,
            spline_order_sin_dec=spline_order_sin_dec,
            **kwargs)


class DataBackgroundI3EnergyPDF(
        I3EnergyPDF,
        IsBackgroundPDF,
):
    """This is the IceCube energy background PDF, which gets constructed from
    experimental data. This class is derived from I3EnergyPDF.
    """
    def __init__(
            self,
            data_exp,
            log10_energy_binning,
            sin_dec_binning,
            smoothing_filter=None,
            **kwargs,
    ):
        """Constructs a new IceCube energy background PDF from experimental
        data.

        Parameters
        ----------
        data_exp : instance of DataFieldRecordArray
            The array holding the experimental data. The following data fields
            must exist:

                log_energy : float
                    The logarithm of the reconstructed energy value of the data
                    event.
                sin_dec : float
                    The sine of the reconstructed declination of the data event.

        log10_energy_binning : instance of BinningDefinition
            The binning definition for the binning in log10(E).
        sin_dec_binning : instance of BinningDefinition
            The binning definition for the sin(declination).
        smoothing_filter : instance of SmoothingFilter | None
            The smoothing filter to use for smoothing the energy histogram.
            If None, no smoothing will be applied.
        """
        if not isinstance(data_exp, DataFieldRecordArray):
            raise TypeError(
                'The data_exp argument must be an instance of '
                'DataFieldRecordArray! '
                f'It is of type "{classname(data_exp)}"!')

        data_log10_energy = data_exp['log_energy']
        data_sin_dec = data_exp['sin_dec']
        # For experimental data, the MC and physics weight are unity.
        data_mcweight = np.ones((len(data_exp),))
        data_physicsweight = data_mcweight

        # Create the PDF using the base class.
        super().__init__(
            pmm=None,
            data_log10_energy=data_log10_energy,
            data_sin_dec=data_sin_dec,
            data_mcweight=data_mcweight,
            data_physicsweight=data_physicsweight,
            log10_energy_binning=log10_energy_binning,
            sin_dec_binning=sin_dec_binning,
            smoothing_filter=smoothing_filter,
            **kwargs)


class MCBackgroundI3EnergyPDF(
        I3EnergyPDF,
        IsBackgroundPDF,
):
    """This is the IceCube energy background PDF, which gets constructed from
    monte-carlo data. This class is derived from I3EnergyPDF.
    """
    def __init__(
            self,
            data_mc,
            physics_weight_field_names,
            log10_energy_binning,
            sin_dec_binning,
            smoothing_filter=None,
            **kwargs,
    ):
        """Constructs a new IceCube energy background PDF from monte-carlo
        data.

        Parameters
        ----------
        data_mc : instance of DataFieldRecordArray
            The array holding the monte-carlo data. The following data fields
            must exist:

                log_energy : float
                    The logarithm of the reconstructed energy value of the data
                    event.
                sin_dec : float
                    The sine of the reconstructed declination of the data event.
                mcweight: float
                    The monte-carlo weight of the event.

        physics_weight_field_names : str | list of str
            The name or the list of names of the monte-carlo data fields, which
            should be used as physics event weights. If a list is given, the
            weight values of all the fields will be summed to construct the
            final event physics weight.
        log10_energy_binning : BinningDefinition
            The binning definition for the binning in log10(E).
        sin_dec_binning : BinningDefinition
            The binning definition for the sin(declination).
        smoothing_filter : SmoothingFilter instance | None
            The smoothing filter to use for smoothing the energy histogram.
            If None, no smoothing will be applied.
        """
        if not isinstance(data_mc, DataFieldRecordArray):
            raise TypeError(
                'The data_mc argument must be an instance of '
                'DataFieldRecordArray! '
                f'It is of type "{classname(data_mc)}"!')

        if not issequence(physics_weight_field_names):
            physics_weight_field_names = [physics_weight_field_names]
        if not issequenceof(physics_weight_field_names, str):
            raise TypeError(
                'The physics_weight_field_names argument must be '
                'of type str or a sequence of type str! '
                f'It is of type {classname(physics_weight_field_names)}')

        data_log10_energy = data_mc['log_energy']
        data_sin_dec = data_mc['sin_dec']
        data_mcweight = data_mc['mcweight']

        # Calculate the event weights as the sum of all the given data fields
        # for each event.
        data_physicsweight = np.zeros(len(data_mc), dtype=np.float64)
        for name in physics_weight_field_names:
            if name not in data_mc:
                raise KeyError(
                    f'The field "{name}" does not exist in the MC data!')
            data_physicsweight += data_mc[name]

        # Create the PDF using the base class.
        super().__init__(
            pmm=None,
            data_log10_energy=data_log10_energy,
            data_sin_dec=data_sin_dec,
            data_mcweight=data_mcweight,
            data_physicsweight=data_physicsweight,
            log10_energy_binning=log10_energy_binning,
            sin_dec_binning=sin_dec_binning,
            smoothing_filter=smoothing_filter,
            **kwargs)



# -*- coding: utf-8 -*-

import pickle

from skyllh.core.py import (
    int_cast,
)
from skyllh.cluster.srvclt import (
    Message,
    receive_object_from_socket,
)


class Command(object):
    """Base class for a command. A command has a command string plus optional
    additional data.
    """
    _cmd = ''

    def __init__(self):
        super(Command, self).__init__()

    def as_message(self):
        msg = pickle.dumps(self)
        return Message(msg)

    def send(self, sock):
        self.as_message().send(sock)

    def is_same_as(self, cmd):
        return (cmd._cmd == self._cmd)


class ACK(Command):
    _cmd = 'ACK'

    def __init__(self):
        super(ACK, self).__init__()


class MSG(Command):
    _cmd = 'MSG'

    def __init__(self, msg):
        super(MSG, self).__init__()

        self.msg = msg


class ShutdownCN(Command):
    _cmd = 'SHUTDOWNCN'

    def __init__(self):
        super(ShutdownCN, self).__init__()


class RegisterCN(Command):
    _cmd = 'REGCN'

    def __init__(self, cn_start_time, cn_live_time):
        """Creates a register compute node command.

        Parameters
        ----------
        cn_start_time : int
            The compute node's start time as unix time stamp.
        cn_live_time : int
            The compute node's live time. After that time the CN should be
            considered dead.
        """
        super(RegisterCN, self).__init__()

        self.cn_start_time = cn_start_time
        self.cn_live_time = cn_live_time

    @property
    def cn_start_time(self):
        """The CN's start time as unix time stamp.
        """
        return self._cn_start_time

    @cn_start_time.setter
    def cn_start_time(self, t):
        t = int_cast(
            t,
            'The cn_start_time property must be castable to type int!')
        self._cn_start_time = t

    @property
    def cn_live_time(self):
        """The CN's live time in seconds.
        """
        return self._cn_live_time

    @cn_live_time.setter
    def cn_live_time(self, t):
        t = int_cast(
            t,
            'The cn_live_time property must be castable to type int!')
        self._cn_live_time = t


def receive_command_from_socket(sock, blocksize=2048):
    """Receives a command from the given socket.
    """
    return receive_object_from_socket(sock, blocksize=blocksize)



# -*- coding: utf-8 -*-

"""This module defines the interface and provides particular implementations
of PDF ratio fill methods. For a binned PDF ratio it could happen, that some
bins don't have background information but signal information is available.
Hence, a ratio cannot be computed for those bins. The PDF ratio fill method
specifies how those bins should get filled.
"""

import abc
import numpy as np

from skyllh.core.py import (
    float_cast,
)


class PDFRatioFillMethod(object, metaclass=abc.ABCMeta):
    """Abstract base class to implement a PDF ratio fill method. It can happen,
    that there are empty background bins but where signal could possibly be.
    A PDFRatioFillMethod implements what happens in such cases.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def __call__(
            self,
            ratios,
            sig_pd_h,
            bkg_pd_h,
            sig_mask_mc_covered,
            sig_mask_mc_covered_zero_physics,
            bkg_mask_mc_covered,
            bkg_mask_mc_covered_zero_physics):
        """The __call__ method is supposed to fill the ratio bins (array)
        with the signal / background ratio values. For bins (array elements),
        where the division is undefined, e.g. due to zero background, the fill
        method decides how to fill those bins.

        Note: Bins which have neither signal monte-carlo nor background
              monte-carlo coverage, are undefined about their signal-ness or
              background-ness by construction.

        Parameters
        ----------
        ratios : ndarray of float
            The multi-dimensional array for the final ratio bins. The shape is
            the same as the sig_h and bkg_h ndarrays.
        sig_pd_h : ndarray of float
            The multi-dimensional array (histogram) holding the signal
            probability densities.
        bkg_pd_h : ndarray of float
            The multi-dimensional array (histogram) holding the background
            probability densities.
        sig_mask_mc_covered : ndarray of bool
            The mask array indicating which array elements of sig_pd_h have
            monte-carlo coverage.
        sig_mask_mc_covered_zero_physics : ndarray of bool
            The mask array indicating which array elements of sig_pd_h have
            monte-carlo coverage but don't have physics contribution.
        bkg_mask_mc_covered : ndarray of bool
            The mask array indicating which array elements of bkg_pd_h have
            monte-carlo coverage.
            In case of experimental data as background, this mask indicate where
            (experimental data) background is available.
        bkg_mask_mc_covered_zero_physics : ndarray of bool
            The mask array ndicating which array elements of bkg_pd_h have
            monte-carlo coverage but don't have physics contribution.
            In case of experimental data as background, this mask contains only
            False entries.

        Returns
        -------
        ratios : ndarray
            The array holding the final ratio values.
        """
        return ratios


class Skylab2SkylabPDFRatioFillMethod(PDFRatioFillMethod):
    """This PDF ratio fill method implements the exact same fill method as in
    the skylab2 software named "skylab". It exists just for comparsion and
    backward compatibility reasons. In general, it should not be used, because
    it does not distinguish between bins with MC converage and physics model
    contribution, and those with MC coverage and no physics model contribution!
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.signallike_percentile = 99.

    @property
    def signallike_percentile(self):
        """The percentile of signal-like ratios, which should be taken as the
        ratio value for ratios with no background probability. This percentile
        must be given as a float value in the range [0, 100] inclusively.
        """
        return self._signallike_percentile

    @signallike_percentile.setter
    def signallike_percentile(self, value):
        value = float_cast(
            value,
            'The value for the signallike_percentile property must be castable '
            'to type float!')
        if (value < 0) or (value > 100):
            raise ValueError(
                f'The value "{value}" of the signallike_percentile property '
                'must be in the range [0, 100]!')
        self._signallike_percentile = value

    def __call__(
            self,
            ratio,
            sig_pd_h,
            bkg_pd_h,
            sig_mask_mc_covered,
            sig_mask_mc_covered_zero_physics,
            bkg_mask_mc_covered,
            bkg_mask_mc_covered_zero_physics):
        """Fills the ratio array ``ratio``.
        For more information see the documentation of
        :meth:`skyllh.core.pdfratio_fill.PDFRatioFillMethod.__call__`.
        """
        # Check if we have predicted background for the entire background MC
        # range.
        if np.any(bkg_mask_mc_covered_zero_physics):
            raise ValueError(
                'Some of the background bins have MC coverage but no physics '
                'background prediction. I don\'t know what to do in this case!')

        sig_domain = sig_pd_h > 0
        bkg_domain = bkg_pd_h > 0

        sig_bkg_domain = sig_domain & bkg_domain

        ratio[sig_bkg_domain] = (
            sig_pd_h[sig_bkg_domain] / bkg_pd_h[sig_bkg_domain]
        )

        ratio_value = np.percentile(
            ratio[ratio > 1.], self._signallike_percentile)
        np.copyto(ratio, ratio_value, where=sig_domain & ~bkg_domain)

        return ratio


class MostSignalLikePDFRatioFillMethod(PDFRatioFillMethod):
    """PDF ratio fill method to set the PDF ratio to the most signal like PDF
    ratio for bins, where there is signal but no background coverage.
    """
    def __init__(self, signallike_percentile=99., **kwargs):
        """Creates the PDF ratio fill method object for filling PDF ratio bins,
        where there is signal MC coverage but no background (MC) coverage
        with the most signal-like ratio value.

        Parameters
        ----------
        signallike_percentile : float in range [0., 100.], default 99.
            The percentile of signal-like ratios, which should be taken as the
            ratio value for ratios with no background probability.
        """
        super().__init__(**kwargs)

        self.signallike_percentile = signallike_percentile

    @property
    def signallike_percentile(self):
        """The percentile of signal-like ratios, which should be taken as the
        ratio value for ratios with no background probability. This percentile
        must be given as a float value in the range [0, 100] inclusively.
        """
        return self._signallike_percentile

    @signallike_percentile.setter
    def signallike_percentile(self, value):
        value = float_cast(
            value,
            'The value for the signallike_percentile property must be castable '
            'to type float!')
        if (value < 0) or (value > 100):
            raise ValueError(
                f'The value "{value}" of the signallike_percentile property '
                'must be in the range [0, 100]!')
        self._signallike_percentile = value

    def __call__(
            self,
            ratio,
            sig_pd_h,
            bkg_pd_h,
            sig_mask_mc_covered,
            sig_mask_mc_covered_zero_physics,
            bkg_mask_mc_covered,
            bkg_mask_mc_covered_zero_physics):
        """Fills the ratio array ``ratio``.
        For more information see the documentation of
        :meth:`skyllh.core.pdfratio_fill.PDFRatioFillMethod.__call__`.
        """
        # Check if we have predicted background for the entire background MC
        # range.
        if np.any(bkg_mask_mc_covered_zero_physics):
            raise ValueError(
                'Some of the background bins have MC coverage but no physics '
                'background prediction. I don\'t know what to do in this case!')

        # Fill the bins where we have signal and background MC coverage.
        mask_sig_and_bkg_mc_covered = sig_mask_mc_covered & bkg_mask_mc_covered
        ratio[mask_sig_and_bkg_mc_covered] = (
            sig_pd_h[mask_sig_and_bkg_mc_covered] /
            bkg_pd_h[mask_sig_and_bkg_mc_covered]
        )

        # Calculate the ratio value, which should be used for ratio bins, where
        # we have signal MC coverage but no background MC coverage.
        ratio_value = np.percentile(
            ratio[ratio > 1.], self._signallike_percentile)
        mask_sig_but_notbkg_mc_covered = (
            sig_mask_mc_covered & ~bkg_mask_mc_covered
        )
        np.copyto(ratio, ratio_value, where=mask_sig_but_notbkg_mc_covered)

        return ratio


class MinBackgroundLikePDFRatioFillMethod(PDFRatioFillMethod):
    """PDF ratio fill method to set the PDF ratio to the minimal background like
    value for bins, where there is signal but no background coverage.
    """
    def __init__(self, **kwargs):
        """Creates the PDF ratio fill method object for filling PDF ratio bins,
        where there is signal MC coverage but no background (MC) coverage
        with the minimal background-like ratio value.
        """
        super().__init__(**kwargs)

    def __call__(
            self,
            ratio,
            sig_pd_h,
            bkg_pd_h,
            sig_mask_mc_covered,
            sig_mask_mc_covered_zero_physics,
            bkg_mask_mc_covered,
            bkg_mask_mc_covered_zero_physics):
        """Fills the ratio array ``ratio``.
        For more information see the documentation of
        :meth:`skyllh.core.pdfratio_fill.PDFRatioFillMethod.__call__`.
        """
        # Check if we have predicted background for the entire background MC
        # range.
        if np.any(bkg_mask_mc_covered_zero_physics):
            raise ValueError(
                'Some of the background bins have MC coverage but no physics '
                'background prediction. I don\'t know what to do in this case!')

        # Fill the bins where we have signal and background MC coverage.
        mask_sig_and_bkg_mc_covered = sig_mask_mc_covered & bkg_mask_mc_covered
        ratio[mask_sig_and_bkg_mc_covered] = (
            sig_pd_h[mask_sig_and_bkg_mc_covered] /
            bkg_pd_h[mask_sig_and_bkg_mc_covered]
        )

        # Calculate the minimal background-like value.
        min_bkg_prob = np.min(bkg_pd_h[bkg_mask_mc_covered])

        # Set the ratio using the minimal background probability where we
        # have signal MC coverage but no background (MC) coverage.
        mask_sig_but_notbkg_mc_covered = (
            sig_mask_mc_covered & ~bkg_mask_mc_covered
        )
        ratio[mask_sig_but_notbkg_mc_covered] =\
            sig_pd_h[mask_sig_but_notbkg_mc_covered] / min_bkg_prob

        return ratio



# -*- coding: utf-8 -*-

import logging
import socket

from skyllh.cluster.commands import (
    ACK,
    Command,
    MSG,
    ShutdownCN,
    RegisterCN,
    receive_command_from_socket,
)


class CNRegistryEntry(object):
    """This class provides an registry entry for a compute node. It holds the
    socket to the compute node.
    """
    def __init__(self, sock, addr, port, cn_start_time, cn_live_time):
        super(CNRegistryEntry, self).__init__()

        self.sock = sock
        self.addr = addr
        self.port = port
        self.cn_start_time = cn_start_time
        self.cn_live_time = cn_live_time

    def __del__(self):
        self.sock.close()

    @property
    def key(self):
        """(read-only) The CN's identification key.
        """
        return f'{self.addr:s}:{self.port:d}'

    def send_command(self, cmd):
        if not isinstance(cmd, Command):
            raise TypeError(
                'The cmd argument must be an instance of Command!')
        cmd.send(self.sock)


class MasterNode(object):
    """The MasterNode class provides an entity to run the SkyLLH program as a
    master node, where compute nodes can register to, so the master node can
    distribute work to the compute nodes. The work distribution is handled
    through the MasterNode instance.
    """
    def __init__(self):
        super(MasterNode, self).__init__()

        self.cn_registry = dict()

    @property
    def cn_registry(self):
        """The dictionary with the registered compute nodes.
        """
        return self._cn_registry

    @cn_registry.setter
    def cn_registry(self, d):
        if not isinstance(d, dict):
            raise TypeError(
                'The cn_registry property must be of type dict!')
        self._cn_registry = d

    def clear_cn_registry(self):
        # Close the sockets to all the CNs.
        for (cn_key, cn) in self.cn_registry.items():
            cn.sock.close()

        self.cn_registry = dict()

    def register_compute_nodes(self, n_cn=10, master_port=9999, blocksize=2048):
        logger = logging.getLogger(__name__)

        logger.debug(
            'Clearing the CN registry')
        self.clear_cn_registry()

        logger.debug(
            'Creating server TCP/IP socket')
        serversock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # bind the socket to a public host, and a well-known port
        master_hostname = socket.getfqdn(socket.gethostname())
        logger.debug(
            'Listening on %s:%d with %d simulanious allowed connections',
            master_hostname, master_port, n_cn)
        serversock.bind((master_hostname, master_port))
        serversock.listen(n_cn)

        try:
            while len(self._cn_registry) < n_cn:
                # Accept connections from the compute nodes in order to register
                # them.
                (clientsock, (addr, port)) = serversock.accept()
                logger.debug(
                    'Got inbound connection from %s:%d', addr, port)

                cmd = receive_command_from_socket(
                    clientsock, blocksize=blocksize)
                if not cmd.is_same_as(RegisterCN):
                    raise RuntimeError(
                        'The compute node provided an unknown command '
                        f'"{cmd.as_message().msg}"!')
                ACK().send(clientsock)

                cn = CNRegistryEntry(
                    clientsock, addr, port, cmd.cn_start_time, cmd.cn_live_time)
                self._cn_registry[cn.key] = cn
        finally:
            serversock.close()

    def send_request(self, msg):
        for (cn_key, cn) in self.cn_registry.items():
            cn.send_command(MSG(msg))

    def shutdown_compute_nodes(self):
        """Sends a stop command to all compute nodes.
        """
        for (cn_key, cn) in self._cn_registry.items():
            cn.send_command(ShutdownCN())



# -*- coding: utf-8 -*-

import sys

"""The session module provides global settings for session handling.
"""

# By default SkyLLH will not be in interactive session, i.e. will be in batch
# mode. Hence, progress bars will not be displayed to not screw up the output.
IS_INTERACTIVE_SESSION = False


def enable_interactive_session():
    """Enables interactive session mode.
    """
    global IS_INTERACTIVE_SESSION

    IS_INTERACTIVE_SESSION = True


def disable_interactive_session():
    """Disables interactive session mode.
    """
    global IS_INTERACTIVE_SESSION

    IS_INTERACTIVE_SESSION = False


def is_interactive_session():
    """Checks whether the current session is interactive (True) or not (False).

    Returns
    -------
    check : bool
        True if the current SkyLLH session is interactive, False otherwise.
    """
    return IS_INTERACTIVE_SESSION


def is_python_interpreter_in_interactive_mode():
    """Checks if the Python interpreter is in interactive mode.

    Returns
    -------
    check : bool
        True if the Python interpreter is in interactive mode, False otherwise.
    """
    return bool(getattr(sys, 'ps1', sys.flags.interactive))



# -*- coding: utf-8 -*-

"""This file contains the global configuration dictionary, together with some
convenience utility functions to set different configuration settings.
"""

import copy
import os.path
import sys

from astropy import (
    units,
)

from typing import (
    Any,
    Dict,
)

from skyllh.core import (
    tool,
)
from skyllh.core.datafields import (
    DataFieldStages as DFS,
)
from skyllh.core.py import (
    classname,
)


_BASECONFIG = {
    'multiproc': {
        # The number of CPUs to use for functions that allow multi-processing.
        # If this setting is set to an int value in the range [1, N] this
        # setting will be used if a function's local ncpu setting is not
        # specified.
        'ncpu': None,
    },
    'debugging': {
        # The default log format.
        'log_format': (
            '%(asctime)s %(processName)s %(name)s %(levelname)s: '
            '%(message)s'),
        # Flag if detailed debug log messages, i.e. trace log messages, should
        # get generated. This is good for debugging but bad for performance.
        'enable_tracing': False,
    },
    'project': {
        # The project's working directory.
        'working_directory': '.',
    },
    'repository': {
        # A base path of repository datasets.
        'base_path': None,
        'download_from_origin': True,
    },
    'units': {
        # Definition of the internal units to use. These must match with the
        # units of the monte-carlo data files.
        'internal': {
            'angle': units.radian,
            'energy': units.GeV,
            'length': units.cm,
            'time': units.s,
        },
        'defaults': {
            # Definition of default units used for fluxes.
            'fluxes': {
                'angle': units.radian,
                'energy': units.GeV,
                'length': units.cm,
                'time': units.s,
            }
        }
    },
    'datafields': {
        'run': DFS.ANALYSIS_EXP,
        'ra': DFS.ANALYSIS_EXP,
        'dec': DFS.ANALYSIS_EXP,
        'ang_err': DFS.ANALYSIS_EXP,
        'time': DFS.ANALYSIS_EXP,
        'log_energy': DFS.ANALYSIS_EXP,
        'true_ra': DFS.ANALYSIS_MC,
        'true_dec': DFS.ANALYSIS_MC,
        'true_energy': DFS.ANALYSIS_MC,
        'mcweight': DFS.ANALYSIS_MC,
    },
    # Flag if specific calculations in the core module can be cached.
    'caching': {
        'pdf': {
            'MultiDimGridPDF': False,
        }
    }
}


class Config(
        dict,
):
    """This class, derived from dict, holds the a local configuration state.
    """

    def __init__(
            self,
    ) -> None:
        """Initializes a new Config instance holding the base configuration.
        """
        super().__init__(copy.deepcopy(_BASECONFIG))

    @classmethod
    @tool.requires('yaml')
    def from_yaml(
            cls,
            pathfilename: str,
    ):
        """Creates a new instance of Config holding the base configuration and
        updated by the configuration items contained in the yaml file using the
        :meth:`dict.update` method.

        Parameters
        ----------
        pathfilename: str | None
            Path and filename to the yaml file containing the to-be-updated
            configuration items.
            If set to ``None``, nothing is done.

        Returns
        -------
        cfg : instance of Config
            The instance of Config holding the base configuration and updated by
            the configuration given in the yaml file.
        """
        cfg = cls()

        if pathfilename is None:
            return cfg

        yaml = tool.get('yaml')

        user_config_dict = yaml.load(
            open(pathfilename),
            Loader=yaml.SafeLoader)
        cfg.update(user_config_dict)

        return cfg

    @classmethod
    def from_dict(
            cls,
            user_dict: Dict[str, Any],
    ):
        """Creates a new instance of Config holding the base configuration and
        updated by the given configuration dictionary using the
        :meth:`dict.update` method.

        Parameters
        ----------
        user_dict: dict
            The dictionary containing the to-be-updated configuration items.

        Returns
        -------
        cfg : instance of Config
            The instance of Config holding the base configuration and updated by
            the given configuration dictionary.
        """
        cfg = cls()

        cfg.update(user_dict)

        return cfg

    @property
    def is_tracing_enabled(self):
        """``True``, if tracing mode is enabled, ``False`` otherwise.
        """
        return self['debugging']['enable_tracing']

    def disable_tracing(
            self,
    ):
        """Disables the tracing mode of SkyLLH.

        Returns
        -------
        self : instance of Config
            The updated instance of Config.
        """
        self['debugging']['enable_tracing'] = False

        return self

    def enable_tracing(
            self,
    ):
        """Enables the tracing mode of SkyLLH.

        Returns
        -------
        self : instance of Config
            The updated instance of Config.
        """
        self['debugging']['enable_tracing'] = True

        return self

    def get_wd(
            self,
    ):
        """Retrieves the absolute path to the working directory as configured in
        this configuration.

        Returns
        -------
        wd : str
            The absolute path to the project's working directory.
        """
        wd = os.path.abspath(self['project']['working_directory'])

        return wd

    def set_enable_tracing(
            self,
            flag,
    ):
        """Sets the setting for tracing.

        Parameters
        ----------
        flag : bool
            The flag if tracing should be enabled (``True``) or disabled
            (``False``).

        Returns
        -------
        self : instance of Config
            The updated instance of Config.
        """
        self['debugging']['enable_tracing'] = flag

        return self

    def set_internal_units(
            self,
            angle_unit=None,
            energy_unit=None,
            length_unit=None,
            time_unit=None,
    ):
        """Sets the units used internally to compute quantities. These units
        must match the units used in the monte-carlo files.

        Parameters
        ----------
        angle_unit : instance of astropy.units.UnitBase | None
            The internal unit that should be used for angles.
            If set to ``None``, the unit is not changed.
        energy_unit : instance of astropy.units.UnitBase | None
            The internal unit that should be used for energy.
            If set to ``None``, the unit is not changed.
        length_unit : instance of astropy.units.UnitBase | None
            The internal unit that should be used for length.
            If set to ``None``, the unit is not changed.
        time_unit : instance of astropy.units.UnitBase | None
            The internal unit that should be used for time.
            If set to ``None``, the unit is not changed.

        Returns
        -------
        self : instance of Config
            The updated instance of Config.
        """
        if angle_unit is not None:
            if not isinstance(angle_unit, units.UnitBase):
                raise TypeError(
                    'The angle_unit argument must be an instance of '
                    'astropy.units.UnitBase!')
            self['units']['internal']['angle'] = angle_unit

        if energy_unit is not None:
            if not isinstance(energy_unit, units.UnitBase):
                raise TypeError(
                    'The energy_unit argument must be an instance of '
                    'astropy.units.UnitBase!')
            self['units']['internal']['energy'] = energy_unit

        if length_unit is not None:
            if not isinstance(length_unit, units.UnitBase):
                raise TypeError(
                    'The length_unit argument must be an instance of '
                    'astropy.units.UnitBase!')
            self['units']['internal']['length'] = length_unit

        if time_unit is not None:
            if not isinstance(time_unit, units.UnitBase):
                raise TypeError(
                    'The time_unit argument must be an instance of '
                    'astropy.units.UnitBase!')
            self['units']['internal']['time'] = time_unit

        return self

    def set_ncpu(
            self,
            ncpu,
    ):
        """Sets the global setting for the number of CPUs to use, when
        parallelization is available.

        Parameters
        ----------
        ncpu : int
            The number of CPUs.

        Returns
        -------
        self : instance of Config
            The updated instance of Config.
        """
        self['multiproc']['ncpu'] = ncpu

        return self

    def set_wd(
            self,
            path=None,
    ):
        """Sets the project's working directory configuration variable and adds
        it to the Python path variable.

        Parameters
        ----------
        cfg : instance of Config
            The instance of Config holding the local configuration.
        path : str | None
            The path of the project's working directory. This can be a path
            relative to the path given by ``os.path.getcwd``, the current
            working directory of the program.
            If set to ``None``, the path is taken from the working directory
            setting of the given configuration.

        Returns
        -------
        wd : str
            The absolute path to the project's working directory.
        """
        if path is None:
            path = self['project']['working_directory']

        if self['project']['working_directory'] in sys.path:
            sys.path.remove(self['project']['working_directory'])

        wd = os.path.abspath(path)
        self['project']['working_directory'] = wd
        sys.path.insert(0, wd)

        return wd

    def to_internal_time_unit(
            self,
            time_unit,
    ):
        """Calculates the conversion factor from the given time unit to the
        internal time unit specified by this local configuration.

        Parameters
        ----------
        time_unit : instance of astropy.units.UnitBase
            The time unit from which to convert to the internal time unit.
        """
        internal_time_unit = self['units']['internal']['time']
        factor = time_unit.to(internal_time_unit)

        return factor

    def wd_filename(self, filename):
        """Generates the fully qualified file name under the project's working
        directory of the given file.

        Parameters
        ----------
        filename : str
            The name of the file for which to generate the working directory
            path file name.

        Returns
        -------
        pathfilename : str
            The generated fully qualified path file name of ``filename`` with
            the project's working directory prefixed.
        """
        pathfilename = os.path.join(self.get_wd(), filename)

        return pathfilename


class HasConfig(
        object,
):
    """Classifier class defining the cfg property. Classes that derive from
    this class indicate, that they hold an instance of Config.
    """

    def __init__(
            self,
            cfg,
            *args,
            **kwargs,
    ):
        """Creates a new instance having the property ``cfg``.

        Parameters
        ----------
        cfg : instance of Config
            The instance of Config holding the local configuration.
        """
        super().__init__(
            *args,
            **kwargs)

        self.cfg = cfg

    @property
    def cfg(self):
        """The instance of Config holding the local configuration.
        """
        return self._cfg

    @cfg.setter
    def cfg(self, c):
        if not isinstance(c, Config):
            raise TypeError(
                'The cfg property must be an instance of Config! '
                f'Currently its type is {classname(c)}!')
        self._cfg = c



# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

"""This module provides classes for defining source catalogs.
"""

from skyllh.core.py import (
    str_cast,
)
from skyllh.core.source_model import (
    SourceModelCollection,
)


class SourceCatalog(
        SourceModelCollection):
    """This class describes a catalog of sources. It is derived from
    SourceModelCollection. A catalog has a name.
    """
    def __init__(
            self,
            name,
            sources=None,
            source_type=None,
            **kwargs):
        """Creates a new source catalog.

        Parameters
        ----------
        name : str
            The name of the catalog.
        sources : sequence of source_type | None
            The sequence of sources this catalog should be initialized with.
        source_type : type | None
            The type of the source class. If set to None (default), the
            default type defined by SourceCollection will be used.
        """
        super().__init__(
            sources=sources,
            source_type=source_type,
            **kwargs)

        self.name = name

    @property
    def name(self):
        """The name of the catalog.
        """
        return self._name

    @name.setter
    def name(self, name):
        name = str_cast(
            name,
            'The name property must be cast-able to type str!')
        self._name = name

    def __str__(self):
        s = f'"{self.name}" {super().__str__()}'
        return s

    def as_SourceModelCollection(self):
        """Creates a SourceModelCollection object for this catalog and
        returns it.

        Returns
        -------
        source_model_collection : instance of SourceModelCollection
            The created instance of SourceModelCollection.
        """
        return SourceModelCollection(
            sources=self.sources,
            source_type=self.source_type)



# -*- coding: utf-8 -*-

import abc

from skyllh.core.config import (
    Config,
    HasConfig,
)
from skyllh.core.dataset import (
    Dataset,
    DatasetData,
)
from skyllh.core.flux_model import (
    FluxModel,
)
from skyllh.core.livetime import (
    Livetime,
)
from skyllh.core.progressbar import (
    ProgressBar,
)
from skyllh.core.py import (
    classname,
    issequence,
    issequenceof,
)
from skyllh.core.types import (
    SourceHypoGroup_t,
)


class DetSigYield(
        object,
        metaclass=abc.ABCMeta,
):
    """This is the abstract base class for a detector signal yield.

    The detector signal yield, Y_s(p_s), is defined as the expected mean
    number of signal events detected by the detector from a given source with
    source parameters p_s.

    To construct a detector signal yield object, four ingredients are
    needed: the dataset holding the monte-carlo data events, a signal flux
    model, the livetime, and a builder instance that knows how to construct
    the actual detector yield in an efficient way.
    In general, the implementation method depends on the detector, the source,
    the flux model with its flux model's signal parameters, and the dataset.
    Hence, for a given detector, source, flux model, and dataset, an appropriate
    implementation method needs to be chosen.
    """
    def __init__(
            self,
            param_names,
            dataset,
            fluxmodel,
            livetime,
            **kwargs,
    ):
        """Constructs a new detector signal yield object. It takes
        the monte-carlo data events, a flux model of the signal, and the live
        time to compute the detector signal yield.

        Parameters
        ----------
        param_names : sequence of str
            The sequence of parameter names this detector signal yield depends
            on. These are either fixed or floating parameters.
        dataset : Dataset instance
            The Dataset instance holding the monte-carlo event data.
        fluxmodel : FluxModel
            The flux model instance. Must be an instance of FluxModel.
        livetime : float | Livetime
            The live-time in days to use for the detector signal yield.
        """
        super().__init__(**kwargs)

        self.param_names = param_names
        self.dataset = dataset
        self.fluxmodel = fluxmodel
        self.livetime = livetime

    @property
    def param_names(self):
        """The tuple of parameter names this detector signal yield instance
        is a function of.
        """
        return self._param_names

    @param_names.setter
    def param_names(self, names):
        if not issequence(names):
            names = [names]
        if not issequenceof(names, str):
            raise TypeError(
                'The param_names property must be a sequence of str '
                'instances! '
                f'Its current type is {classname(names)}.')
        self._param_names = tuple(names)

    @property
    def dataset(self):
        """The Dataset instance, for which this detector signal yield is made
        for.
        """
        return self._dataset

    @dataset.setter
    def dataset(self, ds):
        if not isinstance(ds, Dataset):
            raise TypeError(
                'The dataset property must be an instance of Dataset! '
                f'Its current type is {classname(ds)}.')
        self._dataset = ds

    @property
    def fluxmodel(self):
        """The flux model, which should be used to calculate the detector
        signal yield.
        """
        return self._fluxmodel

    @fluxmodel.setter
    def fluxmodel(self, model):
        if not isinstance(model, FluxModel):
            raise TypeError(
                'The fluxmodel property must be an instance of FluxModel! '
                f'Its current type is {classname(model)}.')
        self._fluxmodel = model

    @property
    def livetime(self):
        """The live-time in days.
        """
        return self._livetime

    @livetime.setter
    def livetime(self, lt):
        if not (isinstance(lt, float) or isinstance(lt, Livetime)):
            raise TypeError(
                'The livetime property must be of type float or an instance '
                'of Livetime! '
                f'Its current type is {classname(lt)}.')
        self._livetime = lt

    @abc.abstractmethod
    def sources_to_recarray(
            self,
            sources,
    ):
        """This method is supposed to convert a (list of) source model(s) into
        a numpy record array that is understood by the detector signal yield
        class.
        This is for efficiency reasons only. This way the user code can
        pre-convert the list of sources into a numpy record array and cache the
        array.
        The fields of the array are detector signal yield implementation
        dependent, i.e. what kind of sources: point-like source, or extended
        source for instance. Because the sources usually don't change their
        position in the sky, this has to be done only once.

        Parameters
        ----------
        sources : SourceModel | sequence of SourceModel
            The source model(s) containing the information of the source(s).

        Returns
        -------
        recarr : numpy record ndarray
            The generated (N_sources,)-shaped 1D numpy record ndarray holding
            the information for each source.
        """
        pass

    @abc.abstractmethod
    def __call__(
            self,
            src_recarray,
            src_params_recarray,
    ):
        """Abstract method to retrieve the detector signal yield for the given
        sources and source parameter values.

        Parameters
        ----------
        src_recarray : (N_sources,)-shaped numpy record ndarray
            The numpy record array containing the information of the sources.
            The required fields of this record array are implementation
            dependent. In the most generic case for a point-like source, it
            must contain the following three fields: ra, dec.
        src_params_recarray : (N_sources,)-shaped numpy record ndarray
            The numpy record ndarray containing the parameter values of the
            sources. The parameter values can be different for the different
            sources.
            The record array must contain two fields for each source parameter,
            one named <name> with the source's local parameter name
            holding the source's local parameter value, and one named
            <name:gpidx> holding the global parameter index plus one for each
            source value. For values mapping to non-fit parameters, the index
            should be negative.

        Returns
        -------
        detsigyield : (N_sources,)-shaped 1D ndarray of float
            The array with the mean number of signal in the detector for each
            given source.
        grads : dict
            The dictionary holding the gradient values for each global fit
            parameter. The key is the global fit parameter index and the value
            is the (N_sources,)-shaped numpy ndarray holding the gradient value
            dY_k/dp_s.
        """
        pass


class DetSigYieldBuilder(
        HasConfig,
        metaclass=abc.ABCMeta,
):
    """Abstract base class for a builder of a detector signal yield. Via the
    ``construct_detsigyield`` method it creates a DetSigYield instance holding
    the internal objects to calculate the detector signal yield.
    """

    def __init__(
            self,
            **kwargs,
    ):
        """Constructor.
        """
        super().__init__(
            **kwargs)

    def assert_types_of_construct_detsigyield_arguments(
            self,
            dataset,
            data,
            shgs,
            ppbar,
    ):

        """Checks the types of the arguments for the ``construct_detsigyield``
        method. It raises errors if the arguments have the wrong type.
        """
        if not isinstance(dataset, Dataset):
            raise TypeError(
                'The dataset argument must be an instance of Dataset! '
                f'Its current type is {classname(dataset)}.')

        if not isinstance(data, DatasetData):
            raise TypeError(
                'The data argument must be an instance of DatasetData! '
                f'Its current type is {classname(data)}.')

        if (not isinstance(shgs, SourceHypoGroup_t)) and\
           (not issequenceof(shgs, SourceHypoGroup_t)):
            raise TypeError(
                'The shgs argument must be an instance of SourceHypoGroup '
                'or a sequence of SourceHypoGroup instances!'
                f'Its current type is {classname(shgs)}.')

        if ppbar is not None:
            if not isinstance(ppbar, ProgressBar):
                raise TypeError(
                    'The ppbar argument must be an instance of ProgressBar! '
                    f'Its current type is {classname(ppbar)}.')

    def get_detsigyield_construction_factory(self):
        """This method is supposed to return a callable with the call-signature

        .. code::

            __call__(
                dataset,
                data,
                shgs,
                ppbar,
            )


        to construct several DetSigYield instances, one for each provided
        source hypo group (i.e. sources and fluxmodel).
        The return value of this callable must be a sequence of DetSigYield
        instances of the same length as the sequence of ``shgs``.

        Returns
        -------
        factory : callable | None
            This default implementation returns ``None``, indicating that a
            factory is not supported by this builder.
        """
        return None

    @abc.abstractmethod
    def construct_detsigyield(
            self,
            dataset,
            data,
            shg,
            ppbar=None,
    ):
        """Abstract method to construct the DetSigYield instance.
        This method must be called by the derived class method implementation
        to ensure the compatibility check of the given flux model with the
        supported flux models.

        Parameters
        ----------
        dataset : instance of Dataset
            The instance of Dataset holding possible dataset specific settings.
        data : instance of DatasetData
            The instance of DatasetData holding the monte-carlo event data.
        shg : instance of SourceHypoGroup
            The instance of SourceHypoGroup (i.e. sources and flux model) for
            which the detector signal yield should be constructed.
        ppbar : instance of ProgressBar | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        detsigyield : instance of DetSigYield
            An instance derived from DetSigYield.
        """
        pass


class NullDetSigYieldBuilder(
        DetSigYieldBuilder):
    """This class provides a dummy detector signal yield builder, which can
    be used for testing purposes, when an actual builder is not required.
    """
    def __init__(
            self,
            cfg=None,
            **kwargs,
    ):
        """Creates a new instance of NullDetSigYieldBuilder.

        Parameters
        ----------
        cfg : instance of Config | None
            The instance of Config holding the local configuration. Since this
            detector signal yield builder does nothing, this argument is
            optional. If not provided the default configuration is used.
        """
        if cfg is None:
            cfg = Config()

        super().__init__(
            cfg=cfg,
            **kwargs)

    def construct_detsigyield(
            self,
            *args,
            **kwargs,
    ):
        """Since this is a dummy detector signal yield builder, calling this
        method will raise a NotImplementedError!
        """
        raise NotImplementedError(
            f'The {classname(self)} detector signal yield builder cannot '
            'actually build a DetSigYield instance!')



# -*- coding: utf-8 -*-

import logging
import os.path
import sys


# Initialize the root logger.
logging.root.setLevel(logging.NOTSET)


def get_logger(
        name):
    """Retrieves the logger with the given name from the Python logging system.

    Parameters
    ----------
    name : str
        The name of the logger.
        Logger hierarchy is defined using dots as separators.

    Returns
    -------
    logger : logging.Logger
        The Logger instance.
    """
    logger = logging.getLogger(name)
    return logger


def setup_logger(
        name,
        log_level):
    """Initializes logger with a given name and a log level.

    Parameters
    ----------
    name : str
        Logger name. Loggers hierarchy is defined using dots as separators.
    log_level : int
        The log level.  The ``logging`` module predefines levels, e.g.
        ``logging.DEBUG``.
    """
    logger = logging.getLogger(name)
    logger.setLevel(log_level)


def setup_console_handler(
        cfg,
        name,
        log_level=None,
        log_format=None,
        stream=None):
    """Initializes `StreamHandler` for a logger with a given name and sets its
    handling level.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    name : str
        Logger name. Loggers hierarchy is defined using dots as separators.
    log_level : int | None
        The log level. The ``logging`` module predefines levels, e.g.
        ``logging.DEBUG``.
        If set to None, the log level of the logger will be used.
    log_format : str | None
        The format of log records in the final output.
        If set to `None`, the log format is taken from the configuration.
    stream : data stream | None
        The stream that the handler should use. Default stream is `sys.stderr`.
    """
    logger = logging.getLogger(name)

    if log_level is None:
        log_level = logger.level

    if log_format is None:
        log_format = cfg['debugging']['log_format']

    if stream is None:
        stream = sys.stderr

    # Create and add `StreamHandler` to the logger.
    sh = logging.StreamHandler(stream=stream)
    sh.setLevel(log_level)
    sh.setFormatter(logging.Formatter(log_format))
    logger.addHandler(sh)


def setup_file_handler(
        cfg,
        name,
        filename,
        log_level=None,
        path=None,
        log_format=None,
        mode='a'):
    """Initializes `FileHandler` for a logger with a given name and sets its
    handling level.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    name : str
        Logger name. Loggers hierarchy is defined using dots as separators.
    log_level : int | None
        The log level. There are predefined levels, e.g. ``logging.DEBUG``.
        If set to None, the log level of the logger will be used.
    filename : str
        The filename of the specified file which is opened and used as the
        stream for logging.
    path : str | None
        The path under which the log file should be stored.
        If set to `None`, the project's working directory will be used.
    log_format : str | None
        The format of log records in the final output.
        If set to `None`, the log format is taken from the configuration.
    mode : str
        File opening mode. Default is 'a' for appending.
    """
    logger = logging.getLogger(name)

    if log_level is None:
        log_level = logger.level

    if path is None:
        path = cfg['project']['working_directory']

    if log_format is None:
        log_format = cfg['debugging']['log_format']

    pathfilename = os.path.join(path, filename)

    # Create and add `FileHandler` to the logger.
    fh = logging.FileHandler(pathfilename, mode=mode)
    fh.setLevel(log_level)
    fh.setFormatter(logging.Formatter(log_format))
    logger.addHandler(fh)



# -*- coding: utf-8 -*-

"""The ``math`` module contains classes for pure mathematical objects.
"""

import abc
from copy import (
    deepcopy,
)
import numpy as np

from skyllh.core.py import (
    classname,
    isproperty,
    issequence,
    issequenceof,
)


class MathFunction(
        object,
        metaclass=abc.ABCMeta):
    """This abstract base class provides an implementation for a mathematical
    function. Such a function has defined parameters, which are implemented as
    class properties. The tuple of parameter names is defined through the
    `param_names` property.
    """

    def __init__(self, **kwargs):
        super(MathFunction, self).__init__(**kwargs)

        self.param_names = ()

    @property
    def param_names(self):
        """The tuple holding the names of the math function's parameters.
        """
        return self._param_names

    @param_names.setter
    def param_names(self, names):
        if not issequence(names):
            names = (names,)
        if not issequenceof(names, str):
            raise TypeError(
                'The param_names property must be a sequence of str!')
        names = tuple(names)
        # Check if all the given names are actual properties of this
        # MathFunction class.
        for name in names:
            if not hasattr(self, name):
                raise KeyError(
                    f'The "{classname(self)}" class does not have an attribute '
                    f'named "{name}"!')
            if not isproperty(self, name):
                raise TypeError(
                    f'The attribute "{classname(self)}" of "{name}" is not a '
                    'property!')
        self._param_names = names

    @property
    @abc.abstractmethod
    def math_function_str(self):
        """The string showing the mathematical function of this MathFunction.
        """
        pass

    def __str__(self):
        """Pretty string representation of this MathFunction instance.
        """
        return self.math_function_str

    def copy(
            self,
            newparams=None):
        """Copies this MathFunction object by calling the copy.deepcopy
        function, and sets new parameters if requested.

        Parameters
        ----------
        newparams : dict | None
            The dictionary with the new parameter values to set, where the
            dictionary key is the parameter name and the dictionary value is the
            new value of the parameter.
        """
        f = deepcopy(self)

        # Set the new parameter values.
        if newparams is not None:
            f.set_params(newparams)

        return f

    def get_param(
            self,
            name):
        """Retrieves the value of the given parameter. It returns ``np.nan`` if
        the parameter does not exist.

        Parameters
        ----------
        name : str
            The name of the parameter.

        Returns
        -------
        value : float | np.nan
            The value of the parameter.
        """
        if name not in self._param_names:
            return np.nan

        value = getattr(self, name)

        return value

    def set_params(
            self,
            pdict):
        """Sets the parameters of the math function to the given parameter
        values.

        Parameters
        ----------
        pdict : dict (name: value)
            The dictionary holding the names of the parameters and their new
            values.

        Returns
        -------
        updated : bool
            Flag if parameter values were actually updated.
        """
        if not isinstance(pdict, dict):
            raise TypeError(
                'The pdict argument must be of type dict!')

        updated = False

        for pname in self._param_names:
            current_value = getattr(self, pname)
            pvalue = pdict.get(pname, current_value)
            if pvalue != current_value:
                setattr(self, pname, pvalue)
                updated = True

        return updated



# -*- coding: utf-8 -*-

import abc

import numpy as np

from skyllh.core.times import (
    TimeGenerator,
)


class DataScramblingMethod(
        object,
        metaclass=abc.ABCMeta,
):
    """Base class for implementing a data scrambling method.
    """

    def __init__(self, **kwargs):
        super().__init__(
            **kwargs)

    @abc.abstractmethod
    def scramble(
            self,
            rss,
            dataset,
            data,
    ):
        """The scramble method implements the actual scrambling of the given
        data, which is method dependent. The scrambling must be performed
        in-place, i.e. it alters the data inside the given data array.

        Parameters
        ----------
        rss : instance of RandomStateService
            The random state service providing the random number
            generator (RNG).
        dataset : instance of Dataset
            The instance of Dataset for which the data should get scrambled.
        data : instance of DataFieldRecordArray
            The DataFieldRecordArray containing the to be scrambled data.

        Returns
        -------
        data : instance of DataFieldRecordArray
            The given DataFieldRecordArray holding the scrambled data.
        """
        pass


class UniformRAScramblingMethod(
        DataScramblingMethod,
):
    r"""The UniformRAScramblingMethod method performs right-ascention scrambling
    uniformly within a given RA range. By default it's (0, 2\pi).

    :note::

        This alters only the ``ra`` values of the data!

    """
    def __init__(
            self,
            ra_range=None,
            **kwargs,
    ):
        r"""Initializes a new RAScramblingMethod instance.

        Parameters
        ----------
        ra_range : tuple | None
            The two-element tuple holding the range in radians within the RA
            values should get drawn from. If set to None, the default (0, 2\pi)
            will be used.
        """
        super().__init__(
            **kwargs)

        self.ra_range = ra_range

    @property
    def ra_range(self):
        """The two-element tuple holding the range within the RA values
        should get drawn from.
        """
        return self._ra_range

    @ra_range.setter
    def ra_range(self, ra_range):
        if ra_range is None:
            ra_range = (0, 2*np.pi)
        if not isinstance(ra_range, tuple):
            raise TypeError(
                'The ra_range property must be a tuple!')
        if len(ra_range) != 2:
            raise ValueError(
                'The ra_range tuple must contain 2 elements!')
        self._ra_range = ra_range

    def scramble(
            self,
            rss,
            dataset,
            data,
    ):
        """Scrambles the given data uniformly in right-ascention.

        Parameters
        ----------
        rss : instance of RandomStateService
            The random state service providing the random number
            generator (RNG).
        dataset : instance of Dataset
            The instance of Dataset for which the data should get scrambled.
        data : instance of DataFieldRecordArray
            The DataFieldRecordArray instance containing the to be scrambled
            data.

        Returns
        -------
        data : instance of DataFieldRecordArray
            The given DataFieldRecordArray holding the scrambled data.
        """
        dt = data['ra'].dtype

        data['ra'] = rss.random.uniform(
            *self.ra_range, size=len(data)).astype(dt, copy=False)

        return data


class TimeScramblingMethod(
        DataScramblingMethod):
    """The TimeScramblingMethod class provides a data scrambling method to
    perform data coordinate scrambling based on a generated time. It draws a
    random time from a time generator and transforms the horizontal (local)
    coordinates into equatorial coordinates using a specified transformation
    function.
    """
    def __init__(
            self,
            timegen,
            hor_to_equ_transform,
            **kwargs,
    ):
        """Initializes a new time scramling method instance.

        Parameters
        ----------
        timegen : instance of TimeGenerator
            The time generator that should be used to generate random MJD times.
        hor_to_equ_transform : callable
            The transformation function to transform coordinates from the
            horizontal system into the equatorial system.

            The call signature must be:

                __call__(azi, zen, mjd)

            The return signature must be: (ra, dec)

        """
        super().__init__(
            **kwargs)

        self.timegen = timegen
        self.hor_to_equ_transform = hor_to_equ_transform

    @property
    def timegen(self):
        """The TimeGenerator instance that should be used to generate random MJD
        times.
        """
        return self._timegen

    @timegen.setter
    def timegen(self, timegen):
        if not isinstance(timegen, TimeGenerator):
            raise TypeError(
                'The timegen property must be an instance of TimeGenerator!')
        self._timegen = timegen

    @property
    def hor_to_equ_transform(self):
        """The transformation function to transform coordinates from the
        horizontal system into the equatorial system.
        """
        return self._hor_to_equ_transform

    @hor_to_equ_transform.setter
    def hor_to_equ_transform(self, transform):
        if not callable(transform):
            raise TypeError(
                'The hor_to_equ_transform property must be a callable object!')
        self._hor_to_equ_transform = transform

    def scramble(
            self,
            rss,
            dataset,
            data,
    ):
        """Scrambles the given data based on random MJD times, which are
        generated from a TimeGenerator instance. The event's right-ascention and
        declination coordinates are calculated via a horizontal-to-equatorial
        coordinate transformation and the generated MJD time of the event.

        Parameters
        ----------
        rss : instance of RandomStateService
            The random state service providing the random number
            generator (RNG).
        dataset : instance of Dataset
            The instance of Dataset for which the data should get scrambled.
        data : instance of DataFieldRecordArray
            The DataFieldRecordArray instance containing the to be scrambled
            data.

        Returns
        -------
        data : instance of DataFieldRecordArray
            The given DataFieldRecordArray holding the scrambled data.
        """
        mjds = self.timegen.generate_times(rss, len(data))

        data['time'] = mjds

        (data['ra'], data['dec']) = self.hor_to_equ_transform(
            data['azi'], data['zen'], mjds)

        return data


class DataScrambler(
        object,
):
    def __init__(
            self,
            method,
            **kwargs,
    ):
        """Creates a data scrambler instance with a given defined scrambling
        method.

        Parameters
        ----------
        method : instance of DataScramblingMethod
            The instance of DataScramblingMethod that defines the method of
            the data scrambling.
        """
        super().__init__(
            **kwargs)

        self.method = method

    @property
    def method(self):
        """The underlaying scrambling method that should be used to scramble
        the data. This must be an instance of the DataScramblingMethod class.
        """
        return self._method

    @method.setter
    def method(self, method):
        if not isinstance(method, DataScramblingMethod):
            raise TypeError(
                'The data scrambling method must be an instance of '
                'DataScramblingMethod!')
        self._method = method

    def scramble_data(
            self,
            rss,
            dataset,
            data,
            copy=False,
    ):
        """Scrambles the given data by calling the scramble method of the
        scrambling method class, that was configured for the data scrambler.
        If the ``inplace_scrambling`` property is set to False, a copy of the
        data is created before the scrambling is performed.

        Parameters
        ----------
        rss : instance of RandomStateService
            The random state service providing the random number generator
            (RNG).
        dataset : instance of Dataset
            The instance of Dataset for which the data should get scrambled.
        data : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the data, which should
            get scrambled.
        copy : bool
            Flag if a copy of the given data should be made before scrambling
            the data. The default is False.

        Returns
        -------
        data : instance of DataFieldRecordArray
            The given DataFieldRecordArray instance with the scrambled data.
            If the ``inplace_scrambling`` property is set to True, this output
            array is the same array as the input array, otherwise it's a new
            array.
        """
        if copy:
            data = data.copy()

        data = self._method.scramble(
            rss=rss,
            dataset=dataset,
            data=data)

        return data
